# This package provides the clyp (curly language yielding python) programming language,
# which is designed to be a simple and intuitive way to write Python code using static typing and curly braces.

__all__ = ['__version__']


__version__ = "1.0.0" # Instead of fancy classes, we'll just use a string for simplicity
